from .code_container import CodeContainer
from .match import Match


class GraphBuilder:
    def __init__(self, g, node_matcher):
        """
        This class performs the operations into the graph g.

        :param g: The graph to modify
        """
        self.g = g
        self.vertices_substitution_dict = {}
        self.edges_substitution_dict = {}
        self.matching_graph = None
        self.matching_code_container = CodeContainer()
        self.match = Match(self.matching_code_container, node_matcher)
        self.update = True
        self.match_info = {}

    def add_graph(self, rhs_graph):
        """
        Adds a graph to self.g

        :param rhs_graph: the graph to add
        :return: itself
        """
        rhs_graph = self.__substitute_names_in_graph(rhs_graph)
        self.g = self.__merge_graphs(self.g, rhs_graph)
        return self

    def set(self, code):
        """
        Executes the code and apply it to the self.g

        :param code: the LIST code to execute
        :return: True/False, depending on the result of the LISP code
        """
        if self.update:
            self.vertices_substitution_dict, self.edges_substitution_dict, self.match_info\
                = self.match.get_variables_substitution_dictionaries(self.g, self.matching_graph)
        try:
            self.matching_graph = self.__apply_code_to_graph(code, self.matching_graph)
        except:
            pass
        try:
            code = self.__substitute_names_in_code(code)
            self.g = self.__apply_code_to_graph(code, self.g)
        except:
            pass
        return True

    def match_graph(self, rhs_graph):
        """
        Sets the graph to match with self.g

        :param rhs_graph: The graph to match
        :return: None
        """
        self.matching_graph = rhs_graph

    def where(self, code_string):
        """
        It sets the LISP code to execute upon matching graphs

        :param code_string: The code to execute
        :return: None
        """
        self.matching_code_container.add_line(code_string)

    def delete_list(self, variables):
        """
        Deletes a list of vertices/edges from self.g

        :param variables: the names of the variables to delete
        :return:
        """
        variables = set(self.__substitute_names_in_list(variables))
        self.update = False
        self.g.delete_vertices(self.g.vs.select(name_in=variables))
        self.g.delete_edges(self.g.es.select(name_in=variables))

    def build(self):
        """
        Return the graph

        :return: self.g
        """
        return self.g

    def get_match_dict(self):
        """

        :return: A dict with the information on how the match went. The keys are:
                    * __RESULT__: True/False
        """

    def build_variables(self, variable_placeholders):
        """
        :param variables: The list of vertices/edges to return
        :return: a dict where the keys are the names of the variables to return,
                 the values are the JSON of the properties of these variables
        """
        variables = self.__substitute_names_in_list(variable_placeholders)
        attributes = {}
        for i, variable in enumerate(variables):
            placeholder_name = variable_placeholders[i]
            try:
                vertices = self.g.vs.select(name=variable)
                attributes[placeholder_name] = vertices[0].attributes()
            except:
                pass
        for i, variable in enumerate(variables):
            placeholder_name = variable_placeholders[i]
            try:
                edges = self.g.es.select(name=variable)
                edge_attr = edges[0].attributes()
                attributes[placeholder_name] = edge_attr
            except:
                pass
        for i, variable in enumerate(variables):
            placeholder_name = variable_placeholders[i]
            try:
                attributes[placeholder_name] = self.match_info[placeholder_name]
            except:
                pass
        return attributes

    # Private

    def __apply_code_to_graph(self, code_string, graph):
        code_container = CodeContainer()
        code_container.add_line(code_string)
        code_container.add_graph_to_namespace(graph)
        code_container.execute()
        code_container.substitute_namespace_into_graph(graph)
        return graph

    def __substitute_names_in_code(self, code):
        for k, v in self.vertices_substitution_dict.items():
            code = code.replace(' ' + k + ' ', ' ' + v + ' ')
        for k, v in self.edges_substitution_dict.items():
            code = code.replace(' ' + k + ' ', ' ' + v + ' ')
        return code

    def __substitute_names_in_graph(self, g):
        if self.update:
            self.vertices_substitution_dict, self.edges_substitution_dict, self.match_info \
                = self.match.get_variables_substitution_dictionaries(self.g, self.matching_graph)
        for i, v in enumerate(g.vs):
            name = v['name']
            try:
                new_name = self.vertices_substitution_dict[name]
                g.vs[i]['name'] = new_name
            except:
                pass
        for i, e in enumerate(g.es):
            name = e['name']
            try:
                new_name = self.edges_substitution_dict[name]
                g.es[i]['name'] = new_name
            except:
                pass
        return g

    def __substitute_names_in_list(self, placeholder_lst):
        return_list = []
        if self.update:
            self.vertices_substitution_dict, self.edges_substitution_dict, self.match_info \
                = self.match.get_variables_substitution_dictionaries(self.g, self.matching_graph)
        for i, placeholder_name in enumerate(placeholder_lst):
            name = placeholder_name
            try:
                name = self.vertices_substitution_dict[name]
            except:
                pass
            try:
                name = self.edges_substitution_dict[name]
            except:
                pass
            return_list.append(name)
        return return_list

    def __merge_graphs(self, lhs, rhs):
        for v in rhs.vs:
            lhs.add_vertex(**v.attributes())
        for e in rhs.es:
            lhs.add_edge(rhs.vs[e.tuple[0]]['name'],
                         rhs.vs[e.tuple[1]]['name'],
                         **e.attributes())
        mapping = []
        name_to_index_dict = {}
        mapped_index = 0
        for v in lhs.vs:
            name = v['name']
            try:
                mapping.append(name_to_index_dict[name])
            except:
                name_to_index_dict[name] = mapped_index
                mapping.append(mapped_index)
                mapped_index += 1
        lhs.contract_vertices(mapping=mapping, combine_attrs='first')
        return lhs



import hy


class CodeContainer:
    def __init__(self):
        self.code_strings = []
        self.namespace = {'result': True}

    def add_line(self, string):
        """
        Adds a line to the LISP code to execute

        :param string: The line to add
        :return: None
        """
        self.code_strings.append(string)

    def add_graph_to_namespace(self, graph):
        """
        Adds the variables name to the namespace of the local LISP code

        :param graph: the graph to add to the namespace
        :return: None
        """
        for node in graph.vs:
            attributes = node.attributes()
            self.namespace[node['name']] = attributes
        for node in graph.es:
            attributes = node.attributes()
            self.namespace[node['name']] = attributes

    def execute(self, vertices_substitution_dict={}):
        """
        Executes the code

        :param vertices_substitution_dict: aliases of the variables in the code
        :return: True/False, depending on the result of the code (default is True)
        """
        code = '(setv result True)'
        if len(self.code_strings) == 1:
            code = '(setv result ' + self.code_strings[0] + ')'
        if len(self.code_strings) > 1:
            code = '(setv result (and ' + ' '.join(self.code_strings) + '))'
        code = self.__substitute_names_in_code(code, vertices_substitution_dict)
        x = hy.lex.tokenize(code)
        try:
            hy.importer.hy_eval(x, self.namespace, "__main__")
        except:
            pass
        return self.namespace['result']

    def substitute_namespace_into_graph(self, graph):
        """
        Creates a graph from the local namespace of the code (to be used after the execution of the code)

        :param graph: The graph to use as a recipient of the namespace
        :return: the updated graph
        """
        for key, value in self.namespace.items():
            try:
                nodes = graph.vs.select(name=key)
                for node in nodes:
                    for k, v in value.items():
                        node[k] = v
            except:
                pass
            try:
                nodes = graph.es.select(name=key)
                for node in nodes:
                    for k, v in value.items():
                        node[k] = v
            except:
                pass
        return graph

    def __substitute_names_in_code(self, code, vertices_substitution_dict):
        for k, v in vertices_substitution_dict.items():
            code = code.replace(' ' + k + ' ', ' ' + v + ' ')
        return code



import random
from igraph import Graph


def get_random_name():
    return "dummy" + str(random.randint(1, 10000))


def convert_graph_to_string(g):
    drt_string = ""
    for vertex in g.vs:
        attributes = vertex.attributes()
        attributes.pop('name')
        drt_string += str(attributes)
        drt_string += '('
        drt_string += vertex['name']
        drt_string += ')'
        drt_string += ', '
    for edge in g.es:
        attributes = edge.attributes()
        attributes.pop('name')
        drt_string += str(attributes)
        drt_string += '('
        drt_string += g.vs[edge.tuple[0]]['name']
        drt_string += ','
        drt_string += g.vs[edge.tuple[1]]['name']
        drt_string += ')'
        drt_string += ', '
    drt_string = drt_string[:-2]

    return drt_string


def is_edge(string):
    return string.find(',') != -1


def create_graph_from_string(graph_string, directed=True):
    import ast

    g = Graph(directed=directed)
    line = graph_string.replace(' ', '')
    line = line.replace('\n', '')
    predicates_strings = line.split("),")
    vertices_to_add = []
    edges_to_add = []
    for predicate in predicates_strings:
        predicate = predicate.replace(')', '')
        attributes_str, name_str = predicate.split("(")
        attributes_dict = ast.literal_eval(attributes_str)
        if not is_edge(name_str):
            attributes_dict['name'] = name_str
            vertices_to_add.append(attributes_dict)
        else:
            if 'name' not in attributes_dict:
                attributes_dict['name'] = get_random_name()
            source, target = name_str.split(',')
            edges_to_add.append((source, target, attributes_dict))
    for attributes_dict in vertices_to_add:
        g.add_vertex(**attributes_dict)
    for source, target, attributes_dict in edges_to_add:
        g.add_edge(source, target, **attributes_dict)
    return g



from .aux import create_graph_from_string, convert_graph_to_string
from .node_matcher import StringNodeMatcher
from .graph_builder import GraphBuilder
from .match import MatchException


def convert_special_characters_to_spaces(line):
    line = line.replace('\t', ' ')
    line = line.replace('\n', ' ')
    return line


class GraphDatabase:
    def __init__(self, g, node_matcher=StringNodeMatcher()):
        """
        This class interprets the commands translates them into operations on a graph by calling GraphBuilder().
        It accepts a graph as an argument and performs operations onto it.

        :param g: The graph to perform operations onto
        :param node_matcher: The class that decides if two nodes match
        """
        self.g = g
        self.node_matcher = node_matcher
        self.action_list = ['MATCH', 'CREATE', 'DELETE', 'RETURN', 'SET', 'WHERE',
                            'match', 'create', 'delete', 'return', 'set', 'where']
        self.action_dict = {'MATCH': self.__match,
                            'CREATE': self.__create,
                            'DELETE': self.__delete,
                            'SET': self.__set,
                            'WHERE': self.__where,
                            'match': self.__match,
                            'create': self.__create,
                            'delete': self.__delete,
                            'set': self.__set,
                            'where': self.__where,
                            }

    def query(self, string):
        """
        This method performs the operations onto self.g

        :param string: The list of operations to perform. The sequences of commands should be separated by a semicolon
                       An example might be
                         CREATE {'tag': 'PERSON', 'text': 'joseph'}(v1), {'relation': 'LIVES_AT'}(v1,v2),
                                {'tag': 'PLACE', 'text': 'London'}(v2)
                         MATCH {}(_a), {'relation': 'LIVES_AT'}(_a,_b), {}(_b)
                           WHERE (= (get _a "text") "joseph")
                         RETURN _a,_b;

        :return: If the RETURN command is called with a list of variables names, a list of JSON with
                 the corresponding properties is returned. If the RETURN command is used alone, a list with the entire
                 graph is returned. Otherwise it returns an empty list
        """
        repeat_n_times = self.__determine_how_many_times_to_repeat_query(string)
        lines = self.__get_command_lines(string)
        return_list = []
        for line in lines:
            lst = self.__query_n_times(line, repeat_n_times)
            if lst and lst[0]:
                return_list = lst
        return return_list

    def get_graph(self):
        return self.g

    # Private

    def __query_n_times(self, line, n):
        builder = GraphBuilder(self.g, self.node_matcher)
        rows = []
        for _ in range(n):
            try:
                result = self.__query_with_builder(line, builder)
                rows.append(result)
                if not result:
                    break
            except MatchException:
                break
            for k, v in result.items():
                try:
                    builder.where('(not (= ' + k + ' ' + v['name'] + '))')
                except:
                    pass
        return rows

    def __query_with_builder(self, string, builder):
        """
        Uses the builder in the argument to modify the graph, according to the commands in the string

        :param string: The single query to the database
        :return: The result of the RETURN operation
        """
        action_graph_pairs = self.__get_action_graph_pairs_from_query(string)
        for action, graph_str in action_graph_pairs:
            if action == 'RETURN' or action == '':
                return self.__return(graph_str, builder)
            try:
                self.action_dict[action](graph_str, builder)
            except MatchException:
                break
        return {}
            
    def __get_action_graph_pairs_from_query(self, query):
        """
        Splits the query into command/argument pairs, for example [("MATCH","{}(_a))", ("RETURN","_a")]

        :param query: The string with the list of commands
        :return: the command/argument pairs
        """
        import re

        query = convert_special_characters_to_spaces(query)
        graph_list = re.split('|'.join(self.action_list), query)
        query_list_positions = [query.find(graph) for graph in graph_list]
        query_list_positions = query_list_positions
        query_list_positions = query_list_positions
        action_list = [query[query_list_positions[i] + len(graph_list[i]):query_list_positions[i + 1]].strip()
                       for i in range(len(graph_list) - 1)]
        graph_list = graph_list[1:]
        return zip(action_list, graph_list)

    def __match(self, graph_str, builder):
        graph = create_graph_from_string(graph_str)
        builder.match_graph(graph)

    def __create(self, graph_str, builder):
        graph = create_graph_from_string(graph_str)
        builder.add_graph(graph)

    def __delete(self, graph_str, builder):
        variables = [v for v in graph_str.strip().replace(' ', '').split(',') if v]
        builder.delete_list(variables)

    def __return(self, graph_str, builder):
        variables = [v for v in graph_str.strip().replace(' ', '').split(',') if v]
        if not variables:
            return {'GRAPH': convert_graph_to_string(builder.build())}
        return builder.build_variables(variables)

    def __set(self, graph_str, builder):
        builder.set(graph_str)
        return True

    def __where(self, graph_str, builder):
        builder.where(graph_str)
        return True

    def __get_command_lines(self, string):
        lines = []
        for line in string.split('\n'):
            if not line.strip() or line.strip()[0] == '#':
                continue
            lines.append(line)
        lines = '\n'.join(lines).split(';')
        return lines

    def __determine_how_many_times_to_repeat_query(self, query_string):
        repeat_n_times = len(self.g.vs)
        if query_string.find('CREATE') != -1:
            repeat_n_times = 1
        return repeat_n_times



class MatchException(Exception):
    def __init__(self):
        pass


class Match:
    def __init__(self, matching_code_container, node_matcher):
        self.matching_code_container = matching_code_container
        self.node_matcher = node_matcher

    def get_variables_substitution_dictionaries(self, lhs_graph, rhs_graph):
        """
        Looks for sub-isomorphisms of rhs into lhs

        :param lhs_graph: The graph to look sub-isomorphisms into (the bigger graph)
        :param rhs_graph: The smaller graph
        :return: The list of matching names
        """
        if not rhs_graph:
            return {}, {}, {}
        return self.__collect_variables_that_match_graph(lhs_graph, rhs_graph)

    def __collect_variables_that_match_graph(self, lhs_graph, rhs_graph):
        match_info = {}
        is_match, mapping, _ = lhs_graph.subisomorphic_vf2(other=rhs_graph,
                                                           return_mapping_12=True,
                                                           node_compat_fn=self.__node_compare,
                                                           edge_compat_fn=self.__edge_compare)
        if not is_match:
            raise MatchException()
        vertices_substitution_dict = {}
        edges_substitution_dict = {}
        for lhs, rhs in enumerate(mapping):
            if rhs == -1:
                continue
            lhs_name = lhs_graph.vs[lhs]['name']
            rhs_name = rhs_graph.vs[rhs]['name']
            vertices_substitution_dict[rhs_name] = lhs_name
        for rhs_edge in rhs_graph.es:
            source_target_list = self.__substitute_names_in_list([rhs_graph.vs[rhs_edge.tuple[0]]['name'],
                                                                  rhs_graph.vs[rhs_edge.tuple[1]]['name']],
                                                                 vertices_substitution_dict)
            source_index = lhs_graph.vs.select(name=source_target_list[0])[0].index
            target_index = lhs_graph.vs.select(name=source_target_list[1])[0].index
            lhs_edges = lhs_graph.es.select(_source=source_index, _target=target_index)
            for lhs_edge in lhs_edges:
                lhs_name = lhs_edge['name']
                rhs_name = rhs_edge['name']
                edges_substitution_dict[rhs_name] = lhs_name

        match_info['__RESULT__'] = is_match

        return vertices_substitution_dict, edges_substitution_dict, match_info

    def __substitute_names_in_list(self, lst, substitution_dict):
        for i, v in enumerate(lst):
            name = lst[i]
            try:
                new_name = substitution_dict[name]
                lst[i] = new_name
                continue
            except:
                pass
        return lst

    def __node_compare(self, lhs_graph, rhs_graph,
                       lhs_graph_index, rhs_graph_index):
        lhs_attr = lhs_graph.vs[lhs_graph_index].attributes()
        rhs_attr = rhs_graph.vs[rhs_graph_index].attributes()
        lhs_name = lhs_attr.pop('name')
        rhs_name = rhs_attr.pop('name')

        self.matching_code_container.add_graph_to_namespace(lhs_graph)
        self.matching_code_container.add_graph_to_namespace(rhs_graph)
        if not self.matching_code_container.execute({rhs_name: lhs_name}):
            return False
        rhs_attr = {k: v for k, v in rhs_attr.items() if v}
        if self.node_matcher.left_contains_right(rhs_attr, lhs_attr):
            return True
        return False

    def __edge_compare(self, lhs_graph, rhs_graph,
                       lhs_graph_index, rhs_graph_index):
        lhs_attr = lhs_graph.es[lhs_graph_index].attributes()
        rhs_attr = rhs_graph.es[rhs_graph_index].attributes()
        lhs_name = lhs_attr.pop('name')
        rhs_name = rhs_attr.pop('name')

        self.matching_code_container.add_graph_to_namespace(lhs_graph)
        self.matching_code_container.add_graph_to_namespace(rhs_graph)
        if not self.matching_code_container.execute({rhs_name: lhs_name}):
            return False
        rhs_attr = {k: v for k, v in rhs_attr.items() if v}
        if self.node_matcher.left_contains_right(rhs_attr, lhs_attr):
            return True
        return False



from setuptools import setup

setup(name='parvusdb',
      version='0.0.10',
      description='A lightweight in-memory graph database',
      url='http://github.com/fractalego/parvusdb',
      author='Alberto Cetoli',
      author_email='alberto@nlulite.com',
      license='MIT',
      packages=['parvusdb', 'parvusdb.utils'],
      install_requires=[
          'python-igraph',
          'hy',
      ],
      zip_safe=False)



class StringDictionaryMatcher(object):
    """
    Checks whether one dict is contained into another one.
    All the values are matched upon using exact match.
    """

    def _match(self, key, lhs, rhs):
        if lhs != rhs:
            return False
        return True

    def left_contains_right(self, lhs, rhs):
        for key in lhs.keys():
            try:
                lvalue = lhs[key]
                rvalue = rhs[key]
                if not self._match(key, lvalue, rvalue):
                    return False
            except:
                return False
        return True
